@@ -4,22 +4,35 @@
 ## May 3, 2021
 
 """
-=========================================================
-Herbie: Download HRRR and RAP model output from the cloud
-=========================================================
+===============================
+Herbie: Retrieve NWP Model Data
+===============================
 
 Herbie is your model output download assistant with a mind of his own!
 Herbie might look small on the outside, but he has a big heart on the 
 inside and will get you to the
-`finish line <https://www.youtube.com/watch?v=4XWufUZ1mxQ&t=189s>`_ 🏁.
+`finish line <https://www.youtube.com/watch?v=4XWufUZ1mxQ&t=189s>`_.
+Happy racing! 🏎🏁
 
-With Herbie's API, you can download High-Resolution Rapid Refresh (HRRR)
-HRRR-Alaska, and Rapid Refresh (RAP) model GRIB2 files from different 
-archive sources. Supports subsetting of GRIB2 files by individual GRIB
-messages (i.e. variable and level) if the index (.idx) file exist.
-Herbie looks for model output data from NOMADS, NOAA's Big Data Project 
-partners (Amazon Web Services, Google Cloud Platform, and Microsoft 
-Azure), and the CHPC Pando archive at the University of Utah.
+`📔 Documentation <https://blaylockbk.github.io/Herbie/_build/html/>`_
+
+With Herbie's API, you can search and download GRIB2 model output files
+from different archive sources for the High-Resolution Rapid Refresh 
+(HRRR) HRRR-Alaska, Rapid Refresh (RAP), Global Forecast System (GFS), 
+and others.
+
+Herbie looks for GRIB2 model output data from NOMADS, NOAA's Big Data 
+Project partners (Amazon Web Services, Google Cloud Platform, and 
+Microsoft Azure), and the CHPC Pando archive at the University of Utah.
+
+Herbie supports subsetting of GRIB2 files by individual GRIB
+messages (i.e. variable and level) when the index (.idx) file exist and
+help you open them with xarray/cfgrib.
+
+Herbie is extendable to support other models. Simply create a template
+file in the ``herbie/models`` directory and make a pull-request.
+
+For more details, see https://blaylockbk.github.io/Herbie/_build/html/user_guide/data_sources.html
 
 .. note:: Updates since ``hrrrb``
 
@@ -35,96 +48,36 @@ Azure), and the CHPC Pando archive at the University of Utah.
     - Option to remove grib2 file when reading xarray if didn't already exist locally (don't clutter local disk).
     - Attach index file DataFrame to object if it exists.
     - If full file exists locally, use remote idx file to cURL local file instead of remote. (Can't create idx file locally because wgrib2 not available on windows)
+    - Added GFS data, though it isn't implemented as cleanly as HRRR or RAP
+    - Renamed 'field' to 'product'
+    - ✨ Moved the source URL templates to their own classes in the **models** folder
+    - Renamed GitHub repository to Herbie (from HRRR_archive_download)
+    - Added RRFS, NBM, GFS, RAP as models Herbie can download
+    - Reworked read_idx() to support index files with additional info (specifically for the NBM).
     - TODO: Rename 'searchString' to 'subset' (and rename subset function)
-    - TODO: Add NCEI as a source for the RAP data?? URL is complex.
-    - TODO: Create .idx file if wgrib2 is installed (linux only)
-
-HRRR and RAP Data Sources
--------------------------
-Real-time HRRR and RAP data is available from NOMADS. At the end of 
-2020, NOAA started pushing the HRRR and RAP data to the Google Cloud 
-Platform and Amazon Web Services as part of the NOAA Big Data Progam.
-These archives were backfilled to previous years as far back as 
-July 30, 2014. A archive used for research purposes also exists 
-on the University of Utah CHPC Pando Archive System.
-
-.. note::
-    A paper about archiving model data in the cloud.
-
-    Blaylock B., J. Horel and S. Liston, 2017: Cloud Archiving and Data
-    Mining of High Resolution Rapid Refresh Model Output. Computers and
-    Geosciences. 109, 43-50. https://doi.org/10.1016/j.cageo.2017.08.005
-
-The Herbie module enables you to easily download HRRR and RAP data 
-between these different data sources wherever the data you are 
-interested in is available. The different download sources and the 
-download priority order are listed below:
-
-#. AWS: Amazon Web Services
-    - https://noaa-hrrr-bdp-pds.s3.amazonaws.com/
-    - Available from July 30, 2014 to Present.
-    - Slight latency for real-time products.
-    - Not all GRIB2 files have an .idx files.
-    - Has all nat, subh, prs, and sfc files for all forecast hours.
-    - Some data may be missing.
-
-#. NOMADS: NOAA Operational Model Archive and Distribution System
-    - https://nomads.ncep.noaa.gov/
-    - Available for today's and yesterday's runs
-    - Real-time data.
-    - Original data source. All available data included.
-    - Download limits.
-    - Includes GRIB2 .idx for all GRIB2 files.
-
-#. Google: Google Cloud Platform Earth
-    - https://console.cloud.google.com/storage/browser/high-resolution-rapid-refresh
-    - Available from July 30, 2014 to Present.
-    - Slight latency for real-time products.
-    - Does not have GRIB2 .idx files before September 17, 2018.
-    - Has all original data including nat, subh, prs, and sfc files 
-      for all forecast hours.
-
-#. Azure: Microsoft Azure
-    - https://github.com/microsoft/AIforEarthDataSets/blob/main/data/noaa-hrrr.md
-    - Only recent HRRR and RAP data?
-    - Subset of HRRR and RAP data?
-
-#. Pando: The University of Utah HRRR archive
-    - http://hrrr.chpc.utah.edu/
-    - Research archive. Older files being removed.
-    - A subset of prs and sfc files.
-    - Contains .idx file for every GRIB2 file.
-
-.. note:: **Default Download Priority Rational** 
-    Most often, a user will request model output from a recent
-    past or earlier run. Past data is archived at one of NOAA's Big 
-    Data parters. NOMADS has the most recent model output available, 
-    but they also throttle the download speed and will block users who
-    violate their usage agreement and download too much data within
-    an hour. To prevent being blocked by NOMADS, the default is to 
-    first look for data on AWS. If the data requested is within the
-    last few hours and was not found on AWS, then Herbie will look
-    for the data at NOMADS. If you really want to download data 
-    from NOMADS and not a Big Data Project partner, change the
-    priority order so that NOMADS is first, (e.g.,
-    ``priority=['nomads', ...]``).
+    - TODO: Create .idx file if wgrib2 is installed (linux only) when index file doesn't exist
+    - TODO: add `idx_to_df()` and `df_to_idx()` methods.
+    - TODO: clean up document examples. It's kind of scattered now.
 
 """
 import os
-from pathlib import Path
 import warnings
 from datetime import datetime, timedelta
-
 import urllib.request
 import requests
 import cfgrib
 import pandas as pd
 
+# Path imported from __init__ because it has my custom `expand()` method
+from . import Path
+
 # NOTE: These values are set in the config file at 
 # ~/.config/herbie/config.cfg and are read in from the __init__ file.
 from . import _default_save_dir
 from . import _default_priority
 
+import herbie.models as models_template
+
 def _searchString_help():
     """Help/Error Message for `searchString`"""
     msg = [
@@ -157,11 +110,10 @@ def _searchString_help():
 class Herbie:
     """Find model output file location based on source priority."""
 
-    def __init__(self, DATE, fxx=0, *, 
-                 model='hrrr', field='sfc',
+    def __init__(self, date=None, *, valid_date=None,
+                 model='hrrr', fxx=0, product=None, member=1,
                  priority=_default_priority,
                  save_dir=_default_save_dir,
-                 DATE_is_valid_time=False,
                  overwrite=False,
                  verbose=True):
         """
@@ -169,32 +121,36 @@ class Herbie:
         
         Parameters
         ----------
-        DATE : pandas-parsable datetime
-            Model initialization datetime if 
-            ``DATE_is_valid_time=False`` (default) or forecast valid 
-            datetime if ``DATE_is_valid_time=True``.
+        date : pandas-parsable datetime
+            *Model initialization datetime*. 
+            If None, then must set ``valid_date``.
+        valid_date : pandas-parsable datetime
+            Model valid datetime. Must set when ``date`` is None.
         fxx : int
             Forecast lead time in hours. Available lead times depend on
-            the model type and model version. Range from 0 to 15, 18, 
-            36, or 48 (model and run dependant).
-        model : {'hrrr', 'hrrrak', 'rap'}
-            Model type. 
+            the model type and model version. Range is model and run 
+            dependant.
+        model : {'hrrr', 'hrrrak', 'rap', 'gfs', 'gfs_wave', 'rrfs', etc.}
+            Model name as defined in the models template folder. CASE INSENSITIVE 
+            Some examples:
             - ``'hrrr'`` HRRR contiguous United States model
             - ``'hrrrak'`` HRRR Alaska model (alias ``'alaska'``)
             - ``'rap'`` RAP model
-        field : {'sfc', 'prs', 'nat', 'subh'}
-            Output variable field file type. In the HRRR model, the
-            different field files have certain variables at levels. 
-            Not required for the RAP model as all variables are in the 
-            same file.
+        product : {'sfc', 'prs', 'nat', 'subh'}
+            Output variable product file type. If not specified, will 
+            use first product in model template file. CASE SENSITIVE.
+            For example, the HRRR model has these products:
             - ``'sfc'`` surface fields
             - ``'prs'`` pressure fields
             - ``'nat'`` native fields
             - ``'subh'`` subhourly fields
+        member : None or int
+            Some ensemble models (e.g. the future RRFS) will need to 
+            specify an ensemble member.
         priority : list or str
             List of model sources to get the data in the order of 
-            download priority. The available data sources and default
-            priority order are listed below.
+            download priority. CASE INSENSITIVE. Some example data 
+            sources and the default priority order are listed below.
             - ``'aws'`` Amazon Web Services (Big Data Program)
             - ``'nomads'`` NOAA's NOMADS server
             - ``'google'`` Google Cloud Platform (Big Data Program)
@@ -204,30 +160,53 @@ class Herbie:
         save_dir : str or pathlib.Path
             Location to save GRIB2 files locally. Default save directory
             is set in ``~/.config/herbie/config.cfg``.
-        DATE_is_valid_time : bool
-            If True, DATE represents the valid time.
-            If False (default), DATE represents the initialization time.
         Overwrite : bool
             If True, look for GRIB2 files even if local copy exists.
             If False (default), use the local copy (still need to find 
             the idx file).
         """
-
-        self.date = pd.to_datetime(DATE)
         self.fxx = fxx
-        self.model = model.lower()
-        self.field = field.lower()
-        self.priority = priority
-        self.save_dir = Path(save_dir).resolve()
-        
-        if DATE_is_valid_time:
-            # The user-supplied DATE represents the forecast valid 
-            # time. Adjust self.date by forecast lead time so it 
-            # represents the model initialization datetime.
-            self.date = self.date - timedelta(hours=self.fxx)
 
+        if date is not None:
+            # User supplied `date`, which is the model initialization datetime.
+            self.date = pd.to_datetime(date)
+            self.valid_date = self.date + timedelta(hours=self.fxx)
+        else:
+            assert valid_date is not None, "`date` or `valid_date` is required."
+            # User supplied `valid_date`, which is the model valid datetime.
+            self.valid_date = pd.to_datetime(valid_date)
+            self.date = self.valid_date - timedelta(hours=self.fxx)
+
+        self.model = model.lower()
+        self.member = member
+        self.product = product
+        
+        self.priority = priority
+        self.save_dir = Path(save_dir).expand()  # yes, the expand is my custom method from __init__
+        self.overwrite = overwrite
+
+        # Get details from the template of the specified model.
+        # This attaches the details from the `models.<model>.template`
+        # class to this Herbie object.
+        # This line is equivalent to `models_template.gfs.template(self)`.
+        # We do it this way because the model name is a variable.
+        # (see https://stackoverflow.com/a/7936588/2383070 for what I'm doing here)
+        getattr(models_template, self.model).template(self)
+        
+        if product is None:
+            # The user didn't specify a product, so lets use the first
+            # product in the model template.
+            self.product = list(self.PRODUCTS)[0]
+            warnings.warn(f'`product` not specified. Will use ["{self.product}"].')
+            # We need to rerun this so the sources have the new product value.
+            getattr(models_template, self.model).template(self)
+        
+        self.product_description = self.PRODUCTS[self.product]
+
+        # Check the user input
         self._validate()
 
+        # Ok, now we are ready to look for the GRIB2 file at each of the remote sources.
         # self.grib is the first existing GRIB2 file discovered.
         # self.idx is the first existing index file discovered.
         self.grib = None
@@ -236,13 +215,23 @@ class Herbie:
         self.idx_source = None
         
         # But first, check if the GRIB2 file exists locally.
-        local_copy = self.get_localPath()
+        local_copy = self.get_localFilePath()
         if local_copy.exists() and not overwrite:
             self.grib = local_copy
             self.grib_source = 'local'
-            # NOTE: We will still get the idx files from a remote.
+            # NOTE: We will still get the idx files from a remote 
+            #       because they aren't stored locally.
         
-        for source in self.priority:            
+        # If priority list is set, we want to search SOURCES in that 
+        # priority order. If priority is None, then search all SOURCES
+        # in the order given by the model template file.
+        # NOTE: A source from the template will not be used if it is not
+        # included in the priority list.
+        if self.priority is not None:
+            self.SOURCES = {key:self.SOURCES[key] for key in self.priority if key in self.SOURCES}
+
+        # Ok, NOW we are ready to search for the remote GRIB2 files...
+        for source in self.SOURCES:            
             if 'pando' in source:
                 # Sometimes pando returns a bad handshake. Pinging
                 # pando first can help prevent that.
@@ -251,7 +240,7 @@ class Herbie:
             # Get the file URL for the source and determine if the 
             # GRIB2 file and the index file exist. If found, store the
             # URL for the GRIB2 file and the .idx file.
-            url = self.get_url(source)
+            url = self.SOURCES[source]
             
             found_grib = False
             found_idx = False
@@ -274,14 +263,15 @@ class Herbie:
                 # Exit loop early if we found both GRIB2 and idx file.
                 break
 
+        # After searching each source, print some info about what we found...
         if verbose: 
             if any([self.grib is not None, self.idx is not None]):
                 print(f'🏋🏻‍♂️ Found',
                       f'\033[32m{self.date:%Y-%b-%d %H:%M UTC} F{self.fxx:02d}\033[m',
-                      f'{self.model.upper()} {self.field}',
+                      f'[{self.model.upper()}] [product={self.product}]',
                       f'GRIB2 file from \033[38;5;202m{self.grib_source}\033[m and',
                       f'index file from \033[38;5;202m{self.idx_source}\033[m.',
-                      f'{" ":100s}')
+                      f'{" ":150s}')
             else:
                 print(f'💔 Did not find a GRIB2 or Index File for',
                       f'\033[32m{self.date:%Y-%b-%d %H:%M UTC} F{self.fxx:02d}\033[m',
@@ -290,7 +280,7 @@ class Herbie:
 
     def __repr__(self):
         """Representation in Notebook"""
-        msg = (f"{self.model.upper()} model {self.field} fields",
+        msg = (f"[{self.model.upper()}] model [{self.product}] product",
                f"run at \033[32m{self.date:%Y-%b-%d %H:%M UTC}",
                f"F{self.fxx:02d}\033[m")
         return ' '.join(msg)
@@ -299,43 +289,45 @@ class Herbie:
         """When class object is printed"""
         msg = [
             f'{self.model=}',
-            f'{self.field=}',
+            f'{self.DETAILS=}',
+            f'{self.DESCRIPTION=}',
+            f'{self.product=}',
             f'{self.fxx=}',
             f'{self.date=}',
             f'{self.priority=}',
+            f'{self.DETAILS=}',
+            f'{self.SOURCES=}'
         ]
         return '\n'.join(msg)
         
     def _validate(self):
-        """Validate the class input arguments"""
-        _models = {'hrrr', 'hrrrak', 'rap'}
-        _fields = {'prs', 'sfc', 'nat', 'subh'}  # Only for HRRR and HRRR-AK
-        
-        assert self.date < datetime.utcnow(), "🔮 Date cannot be in the future."
+        """Validate the Herbie class input arguments"""
         
+        # Accept model alias
         if self.model.lower() == 'alaska':
             self.model = 'hrrrak'
 
-        assert self.fxx in range(49), "Forecast lead time `fxx` is too large"
+        _models = {m for m in dir(models_template) if not m.startswith('__')}
+        _products = set(self.PRODUCTS)
+        
+        assert self.date < datetime.utcnow(), "🔮 `date` cannot be in the future."
         assert self.model in _models, f"`model` must be one of {_models}"
-        if self.model in ['hrrr', 'hrrrak']:
-            assert self.field in _fields, f"`field must be one of {_fields}"
-        else:
-            # field is not needed for RAP model.
-            self.field = ''
+        assert self.product in _products, f"`product` must be one of {_products}"
         
         if isinstance(self.priority, str):
             self.priority = [self.priority]
         
-        self.priority = [i.lower() for i in self.priority]
+        if self.priority is not None:
+            self.priority = [i.lower() for i in self.priority]
 
-        # Don't look for data from NOMADS if requested date is earlier
-        # than yesterday. NOMADS doesn't keep data that old.
-        if 'nomads' in self.priority:
-            yesterday = datetime.utcnow() - timedelta(hours=24)
-            yesterday = pd.to_datetime(f"{yesterday:%Y-%m-%d}")
-            if self.date < yesterday:
-                self.priority.remove('nomads')
+            # Don't look for data from NOMADS if requested date is earlier
+            # than 14 days ago. NOMADS doesn't keep data that old,
+            # (I think this is true of all models).
+            if 'nomads' in self.priority:
+                expired = datetime.utcnow() - timedelta(days=14)
+                expired = pd.to_datetime(f"{expired:%Y-%m-%d}")
+                if self.date < expired:
+                    self.priority.remove('nomads')
     
     def _ping_pando(self):
         """Pinging the Pando server before downloading can prevent a bad handshake."""
@@ -361,83 +353,22 @@ class Herbie:
             url += '.idx'
         return requests.head(url).ok
     
-    def get_url(self, source):
-        """
-        Build the URL for the GRIB2 file for the requested file and data source.
-        
-        Parameters
-        ----------
-        source : {'aws', 'nomads', 'google', 'azure', 'pando', 'pando2'}
-            The data download source for the RAP and HRRR models. 
-            - ``'aws'`` Amazon Web Services (Big Data Program)
-            - ``'nomads'`` NOAA's NOMADS server
-            - ``'google'`` Google Cloud Platform (Big Data Program)
-            - ``'azure'`` Microsoft Azure (Big Data Program)
-            - ``'pando'`` University of Utah Pando Archive (gateway 1)
-            - ``'pando2'`` University of Utah Pando Archive (gateway 2)
-        """
-        # Big Data Program Path Template
-        BDP_path_template = {
-            'rap': f'rap.{self.date:%Y%m%d}/rap.t{self.date:%H}z.awip32f{self.fxx:02d}.grib2',
-            'hrrr': f"hrrr.{self.date:%Y%m%d}/conus/hrrr.t{self.date:%H}z.wrf{self.field}f{self.fxx:02d}.grib2",
-            'hrrrak': f"hrrr.{self.date:%Y%m%d}/alaska/hrrr.t{self.date:%H}z.wrf{self.field}f{self.fxx:02d}.ak.grib2"
-        }
-
-        # Big Data Program Sources (and NOMADS)
-        if source == 'nomads':
-            if self.model == 'rap':
-                base = f'https://nomads.ncep.noaa.gov/pub/data/nccf/com/rap/prod/'
-                path = BDP_path_template[self.model]
-            else:
-                base = 'https://nomads.ncep.noaa.gov/pub/data/nccf/com/hrrr/prod/'
-                path = BDP_path_template[self.model]
-        elif source == 'aws':
-            if self.model == 'rap':
-                base = 'https://noaa-rap-pds.s3.amazonaws.com/'
-                path = BDP_path_template[self.model]
-            else:
-                base = 'https://noaa-hrrr-bdp-pds.s3.amazonaws.com/'
-                path = BDP_path_template[self.model]
-        elif source == 'google':
-            if self.model == 'rap':
-                base = 'https://storage.googleapis.com/rapid-refresh/'
-                path = BDP_path_template[self.model]
-            else:
-                base = 'https://storage.googleapis.com/high-resolution-rapid-refresh/'
-                path = BDP_path_template[self.model]
-        elif source == 'azure':
-            if self.model == 'rap':
-                base = 'https://noaarap.blob.core.windows.net/rap'
-                path = BDP_path_template[self.model]
-            else:
-                base = 'https://noaahrrr.blob.core.windows.net/hrrr/'
-                path = BDP_path_template[self.model]
-        elif source.startswith('pando'):
-            if source[-1] == '2':
-                gateway = 2
-            else:
-                gateway = 1
-            if self.model == 'rap':
-                return None  # No RAP data on Pando
-            else:
-                base = f'https://pando-rgw0{gateway}.chpc.utah.edu/'
-                path = f"{self.model}/{self.field}/{self.date:%Y%m%d}/{self.model}.t{self.date:%H}z.wrf{self.field}f{self.fxx:02d}.grib2"
-
-        return base+path
-
     @property
-    def get_remoteFileName(self):
+    def get_remoteFileName(self, source=None):
         """Predict Remote File Name"""
-        return self.get_url('nomads').split('/')[-1]  # predict name based on nomads source
+        if source is None:
+            source = list(self.SOURCES)[0]
+        return self.SOURCES[source].split('/')[-1]  # predict name based on nomads source
 
     @property
     def get_localFileName(self):
         """Predict Local File Name"""
-        return f"{self.date:%Y%m%d}_{self.get_remoteFileName}"
+        #return f"{self.date:%Y%m%d}_{self.get_remoteFileName}"
+        return self.LOCALFILE
 
-    def get_localPath(self, searchString=None):
+    def get_localFilePath(self, searchString=None):
         """Get path to local file"""
-        outFile = self.save_dir / self.model / f"{self.date:%Y%m%d}" / self.get_localFileName
+        outFile = self.save_dir.expand() / self.model / f"{self.date:%Y%m%d}" / self.get_localFileName
         if searchString is not None:
             # Reassign the index DataFrame with the requested searchString
             self.idx_df = self.read_idx(searchString)
@@ -462,25 +393,31 @@ class Herbie:
             Searches for strings in the index file lines, specifically
             the variable, level, and forecast_time columns.
             Execute ``_searchString_help()`` for examples of a good
-            searchString.
+            searchString. 
 
-            .. include:: ~/searchString_help.rst
+            .. include:: ../user_guide/searchString.rst
         
         Returns
         -------
         A Pandas DataFrame of the index file.
         """
-        assert self.idx is not None, f"No index file for {self.grib}."
+        assert self.idx is not None, f"No index file found for {self.grib}."
+        
+        # Sometimes idx end in ':', other times it doesn't (in some Pando files).
+        # https://pando-rgw01.chpc.utah.edu/hrrr/sfc/20180101/hrrr.t00z.wrfsfcf00.grib2.idx
+        # https://noaa-hrrr-bdp-pds.s3.amazonaws.com/hrrr.20210101/conus/hrrr.t00z.wrfsfcf00.grib2.idx
+        # Sometimes idx has more than the standard messages
+        # https://noaa-nbm-grib2-pds.s3.amazonaws.com/blend.20210711/13/core/blend.t13z.core.f001.co.grib2.idx
         
-        # Open the idx file
         r = requests.get(self.idx)
         assert r.ok, f"Index file does not exist: {self.idx}"   
-
-        read_idx = r.text.split('\n')[:-1]  # last line is empty
-        df = pd.DataFrame([i.split(':') for i in read_idx], 
-                            columns=['grib_message', 'start_byte', 
-                                     'reference_time', 'variable', 
-                                     'level', 'forecast_time', 'none'])
+        df = pd.read_csv(self.idx,
+                         sep=':', 
+                         names=['grib_message', 'start_byte', 
+                                'reference_time', 'variable', 
+                                'level', 'forecast_time', 
+                                '?', '??', '???']
+                        )
 
         # Format the DataFrame
         df['grib_message'] = df['grib_message'].astype(float)  # float because RAP idx files have some decimal grib message numbers
@@ -489,23 +426,28 @@ class Herbie:
         df['start_byte'] = df['start_byte'].astype(int)
         df['end_byte'] = df['start_byte'].shift(-1, fill_value='')
         df['range'] = df.start_byte.astype(str) + '-' + df.end_byte.astype(str)
-        df = df.drop(columns='none')
         df = df.set_index('grib_message')
         df = df.reindex(columns=['start_byte', 'end_byte', 'range', 
                                  'reference_time', 'valid_time', 
-                                 'variable', 'level', 'forecast_time'])
+                                 'variable', 'level', 'forecast_time',
+                                 '?', '??', '???'])
+
+        df = df.dropna(how='all', axis=1)
+        df = df.fillna('')
+
         df.attrs = dict(
+            url=self.idx,
             source=self.idx_source, 
-            description='Index (.idx) file for the GRIB2 file.', 
+            description='Inventory index (.idx) file for the GRIB2 file.', 
             model=self.model, 
-            field=self.field, 
+            product=self.product, 
             lead_time=self.fxx, 
             datetime=self.date
         )
 
         # Filter DataFrame by searchString
         if searchString not in [None, ':']:
-            columns_to_search = df[['variable', 'level', 'forecast_time']].apply(lambda x: ':'.join(x), axis=1)
+            columns_to_search = df.loc[:, 'variable':].apply(lambda x: ':'.join(x).rstrip(':'), axis=1)
             logic = columns_to_search.str.contains(searchString)
             if logic.sum() == 0:
                 print(f"No GRIB messages found. There might be something wrong with {searchString=}")
@@ -513,27 +455,28 @@ class Herbie:
             df = df.loc[logic]
         return df
     
-    def download(self, searchString=None, source=None, *,
-                 save_dir=None, 
-                 overwrite=False, verbose=True,
+    def download(self, searchString=None, *,
+                 source=None, save_dir=None, overwrite=None,
+                 verbose=True,
                  errors='warn'):
         """
         Download file from source.
 
+        Subsetting by variable follows the same principles described here:
+        https://www.cpc.ncep.noaa.gov/products/wesley/fast_downloading_grib.html
+
         Parameters
         ----------
         searchString : str
             If None, download the full file. Else, use regex to subset
             the file by specific variables and levels.
+            .. include:: ../user_guide/searchString.rst
         source : {'nomads', 'aws', 'google', 'azure', 'pando', 'pando2'}
             If None, download GRIB2 file from self.grib2 which is
             the first location the GRIB2 file was found from the 
             priority lists when this class was initialized. Else, you 
             may specify the source to force downloading it from a 
             different location.
-
-            .. include:: ~/searchString_help.rst
-
         save_dir : str or pathlib.Path
             Location to save the model output files.
             If None, uses the default or path specified in __init__.
@@ -571,7 +514,7 @@ class Herbie:
 
             # Download subsets of the file by byte range with cURL.
             for i, (grbmsg, row) in enumerate(self.idx_df.iterrows()):
-                print(f"{i+1:>4g}: GRIB_message={grbmsg:<3g} \033[34m{row.variable}:{row.level}:{row.forecast_time}\033[m")
+                print(f"{i+1:>4g}: GRIB_message={grbmsg:<3g} \033[34m{':'.join(row.values[5:]).rstrip(':')}\033[m")
                 if i == 0:
                     # If we are working on the first item, overwrite the existing file...
                     curl = f'curl -s --range {row.range} {grib_source} > {outFile}'
@@ -584,8 +527,13 @@ class Herbie:
         
         # If the file exists in the localPath and we don't want to 
         # overwrite, then we don't need to download it.
-        outFile = self.get_localPath(searchString=searchString)
-        if outFile.exists() and not overwrite:
+        outFile = self.get_localFilePath(searchString=searchString)
+
+        # This overrides the overwrite specified in __init__
+        if overwrite is not None:
+            self.overwrite = overwrite
+
+        if outFile.exists() and not self.overwrite:
             if verbose: print(f'🌉 Already have local copy --> {outFile}')
             if searchString in [None, ':']:
                 self.local_grib = outFile
@@ -594,13 +542,15 @@ class Herbie:
             return
 
         # Attach the index file to the object (how much overhead is this?)
-        self.idx_df = self.read_idx(searchString)
+        if self.idx is not None:
+            self.idx_df = self.read_idx(searchString)
 
-        # This overwrites the save_dir specified in __init__
+        # This overrides the save_dir specified in __init__
         if save_dir is not None:
-            self.save_dir = save_dir
-        if not hasattr(self.save_dir, 'exists'): 
-            self.save_dir = Path(self.save_dir).resolve()
+            self.save_dir = Path(save_dir).expand()
+        
+        if not hasattr(Path(self.save_dir).expand(), 'exists'): 
+            self.save_dir = Path(self.save_dir).expand()
 
         # Check that data exists
         if self.grib is None:
@@ -619,7 +569,7 @@ class Herbie:
 
         if source is not None:
             # Force download from a specified source and not from first in priority
-            self.grib = self.get_url(source)
+            self.grib = self.SOURCES[source]
             
         # Create directory if it doesn't exist
         if not outFile.parent.is_dir():
@@ -651,7 +601,7 @@ class Herbie:
         download_kwargs = {**dict(overwrite=False), **download_kwargs}
 
         # Download file if local file does not exists
-        local_file = self.get_localPath(searchString=searchString)
+        local_file = self.get_localFilePath(searchString=searchString)
         
         # Only remove grib if it didn't exists before we download it
         remove_grib = not local_file.exists() and remove_grib
@@ -666,13 +616,15 @@ class Herbie:
 
         # Use cfgrib.open_datasets, just in case there are multiple "hypercubes"
         # for what we requested.
-        Hxr = cfgrib.open_datasets(self.get_localPath(searchString=searchString),
+        Hxr = cfgrib.open_datasets(self.get_localFilePath(searchString=searchString),
                                  backend_kwargs=backend_kwargs)
 
         for h in Hxr:
             h.attrs['model'] = self.model
+            h.attrs['product'] = self.product
+            h.attrs['description'] = self.DESCRIPTION
             h.attrs['remote_grib'] = self.grib
-            h.attrs['local_grib'] = self.get_localPath(searchString=searchString)
+            h.attrs['local_grib'] = self.get_localFilePath(searchString=searchString)
 
         if remove_grib:
             # Load the data to memory before removing the file
